# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test for the Group ORM class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions


class TestGroups(AiidaTestCase):
    """Test backend entities and their collections"""

    def setUp(self):
        """Remove all existing Groups."""
        for group in orm.Group.objects.all():
            orm.Group.objects.delete(group.id)

    def test_count(self):
        """Test the `count` method."""
        node_00 = orm.Data().store()
        node_01 = orm.Data().store()
        nodes = [node_00, node_01]

        group = orm.Group(label='label', description='description').store()
        group.add_nodes(nodes)

        self.assertEqual(group.count(), len(nodes))

    def test_creation(self):
        """Test the creation of Groups."""
        node = orm.Data()
        stored_node = orm.Data().store()

        group = orm.Group(label='testgroup')

        with self.assertRaises(exceptions.ModificationNotAllowed):
            # group unstored
            group.add_nodes(node)

        with self.assertRaises(exceptions.ModificationNotAllowed):
            # group unstored
            group.add_nodes(stored_node)

        group.store()

        with self.assertRaises(ValueError):
            # node unstored
            group.add_nodes(node)

        group.add_nodes(stored_node)

        nodes = list(group.nodes)
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].pk, stored_node.pk)

    def test_node_iterator(self):
        """Test the indexing and slicing functionality of the node iterator."""
        node_00 = orm.Data().store()
        node_01 = orm.Data().store()
        node_02 = orm.Data().store()
        node_03 = orm.Data().store()
        nodes = [node_00, node_01, node_02, node_03]

        group = orm.Group(label='label', description='description').store()
        group.add_nodes(nodes)

        # Indexing
        node_indexed = group.nodes[0]
        self.assertTrue(isinstance(node_indexed, orm.Data))
        self.assertIn(node_indexed.uuid, [node.uuid for node in nodes])

        # Slicing
        nodes_sliced = group.nodes[1:3]
        self.assertTrue(isinstance(nodes_sliced, list))
        self.assertEqual(len(nodes_sliced), 2)
        self.assertTrue(all([isinstance(node, orm.Data) for node in nodes_sliced]))
        self.assertTrue(all([node.uuid in set(node.uuid for node in nodes) for node in nodes_sliced]))

    def test_description(self):
        """Test the update of the description both for stored and unstored groups."""
        node = orm.Data().store()

        group_01 = orm.Group(label='testgroupdescription1', description='group_01').store()
        group_01.add_nodes(node)

        group_02 = orm.Group(label='testgroupdescription2', description='group_02')

        # Preliminary checks
        self.assertTrue(group_01.is_stored)
        self.assertFalse(group_02.is_stored)
        self.assertEqual(group_01.description, 'group_01')
        self.assertEqual(group_02.description, 'group_02')

        # Change
        group_01.description = 'new1'
        group_02.description = 'new2'

        # Test that the groups remained in their proper stored state and that
        # the description was updated
        self.assertTrue(group_01.is_stored)
        self.assertFalse(group_02.is_stored)
        self.assertEqual(group_01.description, 'new1')
        self.assertEqual(group_02.description, 'new2')

        # Store group_02 and check that the description is OK
        group_02.store()
        self.assertTrue(group_02.is_stored)
        self.assertEqual(group_02.description, 'new2')

    def test_add_nodes(self):
        """Test different ways of adding nodes."""
        node_01 = orm.Data().store()
        node_02 = orm.Data().store()
        node_03 = orm.Data().store()
        nodes = [node_01, node_02, node_03]

        group = orm.Group(label='test_adding_nodes').store()
        # Single node
        group.add_nodes(node_01)
        # List of nodes
        group.add_nodes([node_02, node_03])

        # Check
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

        # Try to add a node that is already present: there should be no problem
        group.add_nodes(node_01)
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

    def test_remove_nodes(self):
        """Test node removal."""
        node_01 = orm.Data().store()
        node_02 = orm.Data().store()
        node_03 = orm.Data().store()
        node_04 = orm.Data().store()
        nodes = [node_01, node_02, node_03]
        group = orm.Group(label='test_remove_nodes').store()

        # Add initial nodes
        group.add_nodes(nodes)
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

        # Remove a node that is not in the group: nothing should happen
        group.remove_nodes(node_04)
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

        # Remove one orm.Node
        nodes.remove(node_03)
        group.remove_nodes(node_03)
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

        # Remove a list of Nodes and check
        nodes.remove(node_01)
        nodes.remove(node_02)
        group.remove_nodes([node_01, node_02])
        self.assertEqual(set(_.pk for _ in nodes), set(_.pk for _ in group.nodes))

    def test_name_desc(self):
        """Test Group description."""
        group = orm.Group(label='testgroup2', description='some desc')
        self.assertEqual(group.label, 'testgroup2')
        self.assertEqual(group.description, 'some desc')
        self.assertTrue(group.is_user_defined)
        group.store()

        # Same checks after storing
        self.assertEqual(group.label, 'testgroup2')
        self.assertTrue(group.is_user_defined)
        self.assertEqual(group.description, 'some desc')

        # To avoid to find it in further tests
        orm.Group.objects.delete(group.pk)

    def test_delete(self):
        """Test Group deletion."""
        node = orm.Data().store()
        group = orm.Group(label='testgroup3', description='some other desc').store()

        group_copy = orm.Group.get(label='testgroup3')
        self.assertEqual(group.uuid, group_copy.uuid)

        group.add_nodes(node)
        self.assertEqual(len(group.nodes), 1)

        orm.Group.objects.delete(group.pk)

        with self.assertRaises(exceptions.NotExistent):
            # The group does not exist anymore
            orm.Group.get(label='testgroup3')

    def test_rename(self):
        """Test the renaming of a Group."""
        label_original = 'groupie'
        label_changed = 'nogroupie'

        group = orm.Group(label=label_original, description='I will be renamed')

        # Check name changes work before storing
        self.assertEqual(group.label, label_original)
        group.label = label_changed
        self.assertEqual(group.label, label_changed)

        # Revert the name to its original and store it
        group.label = label_original
        group.store()

        # Check name changes work after storing
        self.assertEqual(group.label, label_original)
        group.label = label_changed
        self.assertEqual(group.label, label_changed)

    def test_rename_existing(self):
        """Test that renaming to an already existing name is not permitted."""
        label_group_a = 'group_a'
        label_group_b = 'group_b'

        orm.Group(label=label_group_a, description='I am the Original G').store()

        # Before storing everything should be fine
        group_b = orm.Group(label=label_group_a, description='They will try to rename me')

        # Storing for duplicate group name should trigger UniquenessError
        with self.assertRaises(exceptions.IntegrityError):
            group_b.store()

        # Reverting to unique name before storing
        group_b.label = label_group_b
        group_b.store()

        # After storing name change to existing should raise
        with self.assertRaises(exceptions.IntegrityError):
            group_b.label = label_group_a

    def test_group_uuid_hashing_for_querybuidler(self):
        """QueryBuilder results should be reusable and shouldn't brake hashing."""
        group = orm.Group(label='test_group')
        group.store()

        # Search for the UUID of the stored group
        builder = orm.QueryBuilder()
        builder.append(orm.Group, project=['uuid'], filters={'label': {'==': 'test_group'}})
        [uuid] = builder.first()

        # Look the node with the previously returned UUID
        builder = orm.QueryBuilder()
        builder.append(orm.Group, project=['id'], filters={'uuid': {'==': uuid}})

        # Check that the query doesn't fail
        builder.all()

        # And that the results are correct
        self.assertEqual(builder.count(), 1)
        self.assertEqual(builder.first()[0], group.id)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Serialisation functions for AiiDA types

WARNING: Changing the representation of things here may break people's current saved e.g. things like
checkpoints and messages in the RabbitMQ queue so do so with caution.  It is fine to add representers
for new types though.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from functools import partial
import yaml

import plumpy
from plumpy.utils import AttributesFrozendict

from aiida import common, orm

_NODE_TAG = '!aiida_node'
_GROUP_TAG = '!aiida_group'
_COMPUTER_TAG = '!aiida_computer'
_ATTRIBUTE_DICT_TAG = '!aiida_attributedict'
_PLUMPY_ATTRIBUTES_FROZENDICT_TAG = '!plumpy:attributes_frozendict'
_PLUMPY_BUNDLE = '!plumpy:bundle'


def represent_node(dumper, node):
    """
    Represent a node in YAML

    :param dumper: the dumper to use
    :param node: the node to represent
    :type node: :class:`aiida.orm.nodes.node.Node`
    :return: the representation
    """
    if not node.is_stored:
        raise ValueError('node {}<{}> cannot be represented because it is not stored'.format(type(node), node.uuid))
    return dumper.represent_scalar(_NODE_TAG, u'%s' % node.uuid)


def node_constructor(loader, node):
    """
    Load an aiida node from the yaml representation

    :param loader: the yaml loader
    :param node: the yaml representation
    :return: the aiida node
    :rtype: :class:`aiida.orm.nodes.node.Node`
    """
    yaml_node = loader.construct_scalar(node)
    return orm.load_node(uuid=yaml_node)


def represent_group(dumper, group):
    """
    Represent a group in YAML

    :param dumper: the dumper to use
    :param group: the group to represent
    :type group: :class:`aiida.orm.Group`
    :return: the representation
    """
    if not group.is_stored:
        raise ValueError('group {} cannot be represented because it is not stored'.format(group))
    return dumper.represent_scalar(_GROUP_TAG, u'%s' % group.uuid)


def group_constructor(loader, group):
    """
    Load an aiida group from the yaml representation

    :param loader: the yaml loader
    :param group: the yaml representation
    :return: the aiida group
    :rtype: :class:`aiida.orm.Group`
    """
    yaml_node = loader.construct_scalar(group)
    return orm.load_group(uuid=yaml_node)


def represent_computer(dumper, computer):
    """
    Represent a group in YAML

    :param dumper: the dumper to use
    :param computer: the computer to represent
    :type computer: :class:`aiida.orm.Computer`
    :return: the representation
    """
    if not computer.is_stored:
        raise ValueError('computer {} cannot be represented because it is not stored'.format(computer))
    return dumper.represent_scalar(_COMPUTER_TAG, u'%s' % computer.uuid)


def computer_constructor(loader, computer):
    """
    Load an aiida computer from the yaml representation

    :param loader: the yaml loader
    :param computer: the yaml representation
    :return: the aiida computer
    :rtype: :class:`aiida.orm.Computer`
    """
    yaml_node = loader.construct_scalar(computer)
    return orm.Computer.get(uuid=yaml_node)


class AiiDADumper(yaml.Dumper):
    """
    Custom AiiDA YAML dumper.  Needed so that we don't have to encode each type in the AiiDA graph
    hierarchy separately using a custom representer.
    """

    def represent_data(self, data):
        if isinstance(data, orm.Node):
            return represent_node(self, data)
        if isinstance(data, orm.Computer):
            return represent_computer(self, data)
        if isinstance(data, orm.Group):
            return represent_group(self, data)

        return super(AiiDADumper, self).represent_data(data)


class AiiDALoader(yaml.Loader):
    """AiiDA specific YAML loader"""


def represent_mapping(tag, dumper, mapping):
    """
    Represent an AttributeDict in YAML

    :param tag: the yaml  tag to use
    :param dumper: the dumper to use
    :type dumper: :class:`yaml.dumper.Dumper`
    :param mapping: the attribute dict to represent
    :return: the representation
    """
    return dumper.represent_mapping(tag, mapping)


def mapping_constructor(mapping_type, loader, mapping):
    """
    Construct an AttributeDict from the representation

    :param mapping_type: the class of the mapping to construct, must accept a dictionary as a
        sole constructor argument to be compatible
    :param loader: the yaml loader
    :type loader: :class:`yaml.loader.Loader`
    :param mapping: the attribute dict representation
    :return: the mapping type
    """
    yaml_node = loader.construct_mapping(mapping)
    return mapping_type(yaml_node)


# All the mapping types:

yaml.add_representer(
    common.extendeddicts.AttributeDict, partial(represent_mapping, _ATTRIBUTE_DICT_TAG), Dumper=AiiDADumper)
yaml.add_constructor(
    _ATTRIBUTE_DICT_TAG, partial(mapping_constructor, common.extendeddicts.AttributeDict), Loader=AiiDALoader)

yaml.add_representer(
    AttributesFrozendict, partial(represent_mapping, _PLUMPY_ATTRIBUTES_FROZENDICT_TAG), Dumper=AiiDADumper)
yaml.add_constructor(
    _PLUMPY_ATTRIBUTES_FROZENDICT_TAG, partial(mapping_constructor, AttributesFrozendict), Loader=AiiDALoader)


def represent_bundle(dumper, bundle):
    """
    Represent an AttributeDict in YAML

    :param tag: the yaml  tag to use
    :param dumper: the dumper to use
    :type dumper: :class:`yaml.dumper.Dumper`
    :param bundle: the attribute dict to represent
    :return: the representation
    """
    as_dict = dict(bundle)
    return dumper.represent_mapping(_PLUMPY_BUNDLE, as_dict)


def bundle_constructor(loader, mapping):
    """
    Construct an AttributeDict from the representation

    :param mapping: the class of the mapping to construct, must accept a dictionary as a sole constructor argument to
        be compatible
    :param loader: the yaml loader
    :type loader: :class:`yaml.loader.Loader`
    :param mapping: the attribute dict representation
    :return: the mapping type
    """
    yaml_node = loader.construct_mapping(mapping)
    bundle = plumpy.Bundle.__new__(plumpy.Bundle)
    bundle.update(yaml_node)
    return bundle


yaml.add_representer(plumpy.Bundle, represent_bundle, Dumper=AiiDADumper)
yaml.add_constructor(_PLUMPY_BUNDLE, bundle_constructor, Loader=AiiDALoader)

yaml.add_constructor(_NODE_TAG, node_constructor, Loader=AiiDALoader)
yaml.add_constructor(_GROUP_TAG, group_constructor, Loader=AiiDALoader)
yaml.add_constructor(_COMPUTER_TAG, computer_constructor, Loader=AiiDALoader)


def serialize(data, encoding=None):
    """
    Serialize the given data structure into a string

    The function supports standard data containers such as maps and lists as well as AiiDA nodes which will be
    serialized into strings, before the whole data structure is dumped into a string using YAML.

    :param data: the general data to serialize
    :param encoding: optional encoding for the serialized string
    :return: string representation of the serialized data structure or byte array if specific encoding is specified
    """
    if encoding is not None:
        serialized = yaml.dump(data, encoding=encoding, Dumper=AiiDADumper)
    else:
        serialized = yaml.dump(data, Dumper=AiiDADumper)

    return serialized


def deserialize(serialized):
    """
    Deserialize a string that represents a serialized data structure

    :param serialized: the string representation of serialized data
    :return: the deserialized data structure
    """
    return yaml.load(serialized, Loader=AiiDALoader)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.engine import WorkChain, Process
from aiida.orm import Dict, Bool, Float, Int
from aiida.plugins import CalculationFactory

DEFAULT_INT = 256


class TestWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(TestWorkChain, cls).define(spec)
        spec.input('a', valid_type=Int, help='A pointless integer')
        spec.input('b', valid_type=Float, help='A pointless float')
        spec.input('c.d', valid_type=Bool, help='A pointless boolean')
        spec.input('e', valid_type=Int, default=Int(DEFAULT_INT))


class TestProcessBuilder(AiidaTestCase):

    def setUp(self):
        super(TestProcessBuilder, self).setUp()
        self.assertIsNone(Process.current())
        self.process_class = CalculationFactory('templatereplacer')
        self.builder = self.process_class.get_builder()

    def tearDown(self):
        super(TestProcessBuilder, self).tearDown()
        self.assertIsNone(Process.current())

    def test_process_builder_attributes(self):
        """
        Check that the builder has all the input ports of the process class as attributes
        """
        for name, port in self.process_class.spec().inputs.items():
            self.assertTrue(hasattr(self.builder, name))

    def test_process_builder_set_attributes(self):
        """
        Verify that setting attributes in builder works
        """
        label = 'Test label'
        description = 'Test description'

        self.builder.metadata.label = label
        self.builder.metadata.description = description

        self.assertEquals(self.builder.metadata.label, label)
        self.assertEquals(self.builder.metadata.description, description)

    def test_workchain(self):
        """
        Verify that the attributes of the TestWorkChain can be set but defaults are not there
        """
        builder = TestWorkChain.get_builder()
        builder.a = Int(2)
        builder.b = Float(2.3)
        builder.c.d = Bool(True)
        self.assertEquals(builder, {'a': Int(2), 'b': Float(2.3), 'c': {'d': Bool(True)}, 'metadata': {'options': {}}})

    def test_invalid_setattr_raises(self):
        """
        Verify that __setattr__ cannot be called on a terminal Port
        """
        builder = TestWorkChain.get_builder()
        with self.assertRaises(AttributeError):
            builder.a.b = 3

    def test_dynamic_getters_value(self):
        """
        Verify that getters will return the actual value
        """
        builder = TestWorkChain.get_builder()
        builder.a = Int(2)
        builder.b = Float(2.3)
        builder.c.d = Bool(True)

        # Verify that the correct type is returned by the getter
        self.assertTrue(isinstance(builder.a, Int))
        self.assertTrue(isinstance(builder.b, Float))
        self.assertTrue(isinstance(builder.c.d, Bool))

        # Verify that the correct value is returned by the getter
        self.assertEquals(builder.a, Int(2))
        self.assertEquals(builder.b, Float(2.3))
        self.assertEquals(builder.c.d, Bool(True))

    def test_dynamic_getters_doc_string(self):
        """
        Verify that getters have the correct docstring
        """
        builder = TestWorkChain.get_builder()
        self.assertEquals(builder.__class__.a.__doc__, str(TestWorkChain.spec().inputs['a']))
        self.assertEquals(builder.__class__.b.__doc__, str(TestWorkChain.spec().inputs['b']))
        self.assertEquals(builder.__class__.c.__doc__, str(TestWorkChain.spec().inputs['c']))
        self.assertEquals(builder.c.__class__.d.__doc__, str(TestWorkChain.spec().inputs['c']['d']))

    def test_job_calculation_get_builder_restart(self):
        """
        Test the get_builder_restart method of CalcJobNode class
        """
        from aiida.orm import CalcJobNode

        # Have to set the process type manually, because usually this will be done automatically when the node is
        # instantiated by the process itself. Since we hack it here and instantiate the node directly ourselves we
        # have to set the process type for the restart builder to be able to recreate the process class.
        original = CalcJobNode(
            computer=self.computer, process_type='aiida.calculations:templatereplacer', label='original')
        original.set_option('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        original.set_option('max_wallclock_seconds', 1800)
        original.store()

        builder = original.get_builder_restart()

        self.assertDictEqual(builder.metadata.options, original.get_options())

    def test_code_get_builder(self):
        """
        Test that the get_builder method of Code returns a builder
        where the code is already set.
        """
        from aiida.orm import Code

        code1 = Code()
        # This also sets the code as a remote code
        code1.set_remote_computer_exec((self.computer, '/bin/true'))
        code1.label = 'test_code1'
        code1.set_input_plugin_name('templatereplacer')
        code1.store()

        # Check that I can get a builder
        builder = code1.get_builder()
        self.assertEquals(builder.code.pk, code1.pk)

        # Check that I can set the parameters
        builder.parameters = Dict(dict={})

        # Check that it complains for an unknown input
        with self.assertRaises(AttributeError):
            builder.unknown_parameter = 3

        # Check that it complains if the type is not the correct one
        # (for the templatereplacer, it should be a
        # Dict)
        with self.assertRaises(ValueError):
            builder.parameters = Int(3)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that defines methods required to setup a new AiiDA instance."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

import click

from aiida.cmdline.utils import echo


def create_instance_directories():
    """Create the base directories required for a new AiiDA instance.

    This will create the base AiiDA directory defined by the AIIDA_CONFIG_FOLDER variable, unless it already exists.
    Subsequently, it will create the daemon directory within it and the daemon log directory.
    """
    from .settings import AIIDA_CONFIG_FOLDER, DAEMON_DIR, DAEMON_LOG_DIR, DEFAULT_UMASK

    directory_base = os.path.expanduser(AIIDA_CONFIG_FOLDER)
    directory_daemon = os.path.join(directory_base, DAEMON_DIR)
    directory_daemon_log = os.path.join(directory_base, DAEMON_LOG_DIR)

    umask = os.umask(DEFAULT_UMASK)

    try:
        if not os.path.isdir(directory_base):
            os.makedirs(directory_base)

        if not os.path.isdir(directory_daemon):
            os.makedirs(directory_daemon)

        if not os.path.isdir(directory_daemon_log):
            os.makedirs(directory_daemon_log)
    finally:
        os.umask(umask)


def setup_profile(profile_name, only_config, set_default=False, non_interactive=False, **kwargs):
    """
    Setup an AiiDA profile and AiiDA user (and the AiiDA default user).

    :param profile_name: Profile name
    :param only_config: do not create a new user
    :param set_default: set the new profile as the default
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :param backend: one of 'django', 'sqlalchemy'
    :param email: valid email address for the user
    :param db_host: hostname for the database
    :param db_port: port to connect to the database
    :param db_user: name of the db user
    :param db_pass: password of the db user
    """
    # pylint: disable=superfluous-parens,too-many-locals,too-many-statements,too-many-branches
    from aiida.backends import settings
    from aiida.backends.profile import BACKEND_SQLA, BACKEND_DJANGO
    from aiida.backends.utils import set_backend_type
    from aiida.cmdline.commands import cmd_user
    from aiida.common.exceptions import InvalidOperation
    from aiida.common.setup import create_profile, create_profile_noninteractive
    from aiida.manage.configuration import get_config
    from aiida.manage.manager import get_manager
    from .settings import DEFAULT_AIIDA_USER

    config = get_config()
    manager = get_manager()

    only_user_config = only_config

    # Create the directories to store the configuration files
    create_instance_directories()

    # we need to overwrite this variable for the following to work
    settings.AIIDADB_PROFILE = profile_name

    profile = None

    # ask and store the configuration of the DB
    if non_interactive:
        try:
            profile = create_profile_noninteractive(
                config=config,
                profile_name=profile_name,
                backend=kwargs['backend'],
                email=kwargs['email'],
                db_host=kwargs['db_host'],
                db_port=kwargs['db_port'],
                db_name=kwargs['db_name'],
                db_user=kwargs['db_user'],
                db_pass=kwargs.get('db_pass', ''),
                repo=kwargs['repo'],
                force_overwrite=kwargs.get('force_overwrite', False))
        except ValueError as exception:
            echo.echo_critical("Error during configuation: {}".format(exception))
        except KeyError as exception:
            import traceback
            echo.echo(traceback.format_exc())
            echo.echo_critical(
                "--non-interactive requires all values to be given on the commandline! Missing argument: {}".format(
                    exception.args[0]))
    else:
        try:
            profile = create_profile(config=config, profile_name=profile_name)
        except ValueError as exception:
            echo.echo_critical("Error during configuration: {}".format(exception))

    # Add the created profile and set it as the new default profile
    config.add_profile(profile_name, profile)
    config.set_default_profile(profile_name, overwrite=set_default)
    config.store()

    if only_user_config:
        echo.echo("Only user configuration requested, skipping the migrate command")
    else:
        echo.echo("Executing now a migrate command...")

        backend_choice = profile['AIIDADB_BACKEND']
        if backend_choice == BACKEND_DJANGO:
            echo.echo("...for Django backend")
            backend = manager._load_backend(schema_check=False)  # pylint: disable=protected-access
            backend.migrate()
            set_backend_type(BACKEND_DJANGO)

        elif backend_choice == BACKEND_SQLA:
            echo.echo("...for SQLAlchemy backend")
            backend = manager._load_backend(schema_check=False)  # pylint: disable=protected-access
            backend.migrate()
            set_backend_type(BACKEND_SQLA)

        else:
            raise InvalidOperation("Not supported backend selected.")

    echo.echo("Database was created successfully")
    from aiida import orm

    if not orm.User.objects.find({'email': DEFAULT_AIIDA_USER}):
        echo.echo("Installing default AiiDA user...")
        nuser = orm.User(email=DEFAULT_AIIDA_USER, first_name="AiiDA", last_name="Daemon")
        nuser.is_active = True
        nuser.store()

    email = manager.get_profile().default_user_email
    echo.echo("Starting user configuration for {}...".format(email))
    if email == DEFAULT_AIIDA_USER:
        echo.echo("You set up AiiDA using the default Daemon email ({}),".format(email))
        echo.echo("therefore no further user configuration will be asked.")
    else:
        if non_interactive:
            # Here we map the keyword arguments onto the command line arguments
            # for verdi user configure.  We have to be careful that there the
            # argument names are the same as those int he kwargs dict
            commands = [kwargs['email'], '--non-interactive']

            for arg in ('first_name', 'last_name', 'institution'):
                value = kwargs.get(arg, None)
                if value is not None:
                    commands.extend(('--{}'.format(arg.replace('_', '-')), str(value)))
        else:
            commands = [email]

        # Ask to configure the user
        try:
            # pylint: disable=no-value-for-parameter
            cmd_user.configure(commands)
        except SystemExit:
            # Have to catch this as the configure command will do a sys.exit()
            pass

    echo.echo("Setup finished.")


def delete_repository(profile, non_interactive=True):
    """
    Delete an AiiDA file repository associated with an AiiDA profile.

    :param profile: AiiDA Profile
    :type profile: :class:`aiida.manage.configuration.profile.Profile`
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :type non_interactive: bool
    """
    from six.moves.urllib.parse import urlparse  # pylint: disable=import-error

    pconfig = profile.dictionary
    repo_uri = pconfig.get('AIIDADB_REPOSITORY_URI', '')
    repo_path = urlparse(repo_uri).path
    repo_path = os.path.expanduser(repo_path)

    if not os.path.isabs(repo_path):
        echo.echo_info("Associated file repository '{}' does not exist.".format(repo_path))
        return

    if not os.path.isdir(repo_path):
        echo.echo_info("Associated file repository '{}' is not a directory.".format(repo_path))
        return

    if non_interactive or click.confirm("Delete associated file repository '{}'?\n"
                                        "WARNING: All data will be lost.".format(repo_path)):
        echo.echo_info("Deleting directory '{}'.".format(repo_path))
        import shutil
        shutil.rmtree(repo_path)


def delete_db(profile, non_interactive=True, verbose=False):
    """
    Delete an AiiDA database associated with an AiiDA profile.

    :param profile: AiiDA Profile
    :type profile: :class:`aiida.manage.configuration.profile.Profile`
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :type non_interactive: bool
    :param verbose: if True, print parameters of DB connection
    :type verbose: bool
    """
    from aiida.manage.configuration import get_config
    from aiida.manage.external.postgres import Postgres
    from aiida.common import json

    pdict = profile.dictionary

    postgres = Postgres.from_profile(profile, interactive=not non_interactive, quiet=False)
    postgres.determine_setup()

    if verbose:
        echo.echo_info("Parameters used to connect to postgres:")
        echo.echo(json.dumps(postgres.get_dbinfo(), indent=4))

    db_name = pdict.get('AIIDADB_NAME', '')
    if not postgres.db_exists(db_name):
        echo.echo_info("Associated database '{}' does not exist.".format(db_name))
    elif non_interactive or click.confirm("Delete associated database '{}'?\n"
                                          "WARNING: All data will be lost.".format(db_name)):
        echo.echo_info("Deleting database '{}'.".format(db_name))
        postgres.drop_db(db_name)

    user = pdict.get('AIIDADB_USER', '')
    config = get_config()
    users = [profile.dictionary.get('AIIDADB_USER', '') for profile in config.profiles]

    if not postgres.dbuser_exists(user):
        echo.echo_info("Associated database user '{}' does not exist.".format(user))
    elif users.count(user) > 1:
        echo.echo_info("Associated database user '{}' is used by other profiles "
                       "and will not be deleted.".format(user))
    elif non_interactive or click.confirm("Delete database user '{}'?".format(user)):
        echo.echo_info("Deleting user '{}'.".format(user))
        postgres.drop_dbuser(user)


def delete_from_config(profile, non_interactive=True):
    """
    Delete an AiiDA profile from the config file.

    :param profile: AiiDA Profile
    :type profile: :class:`aiida.manage.configuration.profile.Profile`
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :type non_interactive: bool
    """
    from aiida.manage.configuration import get_config

    if non_interactive or click.confirm("Delete configuration for profile '{}'?\n"
                                        "WARNING: Permanently removes profile from the list of AiiDA profiles.".format(
                                            profile.name)):
        echo.echo_info("Deleting configuration for profile '{}'.".format(profile.name))
        config = get_config()
        config.remove_profile(profile.name)
        config.store()


def delete_profile(profile, non_interactive=True, include_db=True, include_repository=True, include_config=True):
    """
    Delete an AiiDA profile and AiiDA user.

    :param profile: AiiDA profile
    :type profile: :class:`aiida.manage.configuration.profile.Profile`
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :param include_db: Include deletion of associated database
    :type include_db: bool
    :param include_repository: Include deletion of associated file repository
    :type include_repository: bool
    :param include_config: Include deletion of entry from AiiDA configuration file
    :type include_config: bool
    """
    if include_db:
        delete_db(profile, non_interactive)

    if include_repository:
        delete_repository(profile, non_interactive)

    if include_config:
        delete_from_config(profile, non_interactive)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Cached information for each plugin.

This module and the RegistryEntry class should be the sole
location for the (implicit) definition of the registry format.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import


class InvalidPluginEntryError(Exception):
    def __init__(self, msg=''):
        msg = 'Error: Invalid Plugin Registry Entry: {}'.format(msg)
        super(InvalidPluginEntryError, self).__init__(msg)


class RegistryEntry(object):
    """
    Can be created from an entry in the online aiida plugin registry.
    An instance will be created and cached for each entry on update.
    """
    def __init__(self, **kwargs):
        """
        stores registry entry keys and loads the setup_info from the url given in plugin_info
        """
        self.entry_point = kwargs['entry_point']
        self.name = kwargs['name']
        self.pip_url = kwargs.get('pip_url', None)
        self.info_url = kwargs['plugin_info']
        self.load_setup_info(self.info_url)

        '''placeholders'''
        self._entry_points = {}

    def load_setup_info(self, info_url):
        """Load setup kwargs from the link in the registry"""
        from aiida.plugins.utils import load_json_from_url
        self.setup_info = load_json_from_url(info_url)
        required = ['name', 'version', 'author', 'author_email', 'description',
                    'url', 'license']
        for kw in required:
            if kw not in self.setup_info:
                msg = '"{}" is missing the key "{}" in the file in info_url'.format(self.name, kw)
                raise InvalidPluginEntryError(msg)

    @property
    def package_name(self):
        """The name used to import the package"""
        return self.setup_info['name']

    @property
    def version(self):
        """The version of the plugin package"""
        return self.setup_info['version']

    @property
    def entry_points_raw(self):
        """The full entry point spec in setuptools.setup() format"""
        return self.setup_info['entry_points']

    @property
    def entry_point_categories(self):
        """A list of categories for which this plugin exposes entry points"""
        eps = self.entry_points_raw
        return [i.replace('aiida.', '') for i in eps if i.startswith('aiida')]

    @property
    def entry_points(self):
        """A dict of entry point names by category"""
        if not self._entry_points:
            for category in self.entry_point_categories:
                key = 'aiida.' + category
                ep_list = [i[:i.find('=')].strip() for i in self.entry_points_raw[key]]
                self._entry_points[category] = ep_list
        return self._entry_points

    @property
    def cli_apps(self):
        """A list of cli apps installed by this plugin"""
        eps = self.entry_points_raw.get('console_scripts', [])
        return [i[:i.find('=')].strip() for i in eps]

    @property
    def gui_apps(self):
        """A list of GUI apps installed by this plugin"""
        eps = self.entry_points_raw.get('gui_scripts', [])
        return [i[:i.find('=')].strip() for i in eps]

    def install(self, **opts):
        """Call on pip to install the package if not yet installed"""
        installed, new_style, ivers = self.test_installed()
        if installed:
            return True
        if not self.pip_url:
            raise Exception('The plugin author did not provide an automatic install link')
        from pip.commands.install import InstallCommand
        ic = InstallCommand()
        opts, args = ic.parser.parse_args()
        args.append(self.pip_url)
        for k, v in opts.__dict__.items():
            setattr(opts, k, v)
        req_set = ic.run(opts, args)
        req_set.install(opts)
        return self.test_installed()

    def test_installed(self):
        """
        Return wether the plugin is installed

        First, this checks wether the package_name can be imported.
        If not, we know that at least no new style plugin with
        that name is installed.

        Secondly, tests wether all the entry points are currently
        found by the plugin loader. If not, it is considered not
        installed.

        potential failures:
            * loading of the entry points is not tested
            * not properly uninstalled plugins might show up as
                installed if the entry points are still around.
            * it does not distinguish between not installed and
                an old version is installed
        """
        from importlib import import_module
        from pkg_resources import get_distribution
        new_style = False
        installed = True
        iversion = None
        try:
            import_module(self.package_name)
            new_style = True
            iversion = get_distribution(self.package_name).version
        except ImportError:
            new_style = False

        from aiida.plugins.entry_point import get_entry_point_names

        if iversion == self.version or not new_style:
            for cat, ep in self.entry_points.items():
                if not set(ep).issubset(set(get_entry_point_names('aiida.' + cat))):
                    installed = False
        return installed, new_style, iversion

    @property
    def install_requirements(self):
        return self.setup_info.get('install_requires', [])

    @property
    def extras_requirements(self):
        return self.setup_info.get('extras_require', [])

    @property
    def author(self):
        return self.setup_info.get('author')

    @property
    def author_email(self):
        return self.setup_info.get('author_email')

    @property
    def description(self):
        return self.setup_info.get('description')

    @property
    def tags(self):
        return self.setup_info.get('keywords', [])

    @property
    def home_url(self):
        return self.setup_info['url']

    @property
    def installed(self):
        installed, new_style, version = self.test_installed()
        if new_style:
            installed = version or '?'
        return installed

    def format_info(self, **kwargs):
        """
        format and return a datastructure containing all known information about the plugin

        :param format: str, one of [tabulate | dict]
            tabulate: use tabulate to create and return a table of properties as a string
            dict: create a dict of properties
        :param as_str: bool
            format='dict' and as_str=True: return a pretty printed string version of the dict
            format='dict' and as_str=False: return a dictionary
            format='tabulate': as_str is ignored
        """
        fmt = kwargs.pop('format', None)
        as_str = kwargs.pop('as_str', False)
        res = ''
        if fmt == 'tabulate':
            table = []
            table.append(['Name:', self.name])
            table.append(['Version:', self.version])
            table.append(['Installed:', self.installed])
            table.append(['Pip url:', self.pip_url])
            table.append(['Project home:', self.home_url])
            table.append(['Author:', self.author])
            table.append(['Author Email:', self.author_email])
            table.append(['Package:', self.package_name])
            table.append(['Description:', self.description])
            table.append(['Plugins:', ''])
            for category, eps in self.entry_points.items():
                table.append(['', category.capitalize() + ':'])
                table.extend([['', ep] for ep in eps])
                table.append(['', ''])
            if self.cli_apps:
                table.append(['CLI apps:', ''])
                table.extend([['', app] for app in self.cli_apps])
            if self.gui_apps:
                table.append(['GUI apps:', ''])
                table.extend([['', app] for app in self.gui_apps])
            if as_str:
                from tabulate import tabulate
                res = tabulate(table, **kwargs)
            else:
                res = table
        elif fmt == 'dict':
            res = {
                'name': self.name,
                'version': self.version,
                'installed': self.installed,
                'pip url': self.pip_url,
                'project home': self.home_url,
                'author': self.author,
                'author email': self.author_email,
                'package:': self.package_name,
                'description': self.description,
                'plugins': self.entry_points,
                'cli apps': self.cli_apps,
                'gui apps': self.gui_apps
            }
            if as_str:
                import pprint
                res = pprint.pformat(res)
        return res



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Migration after the provenance redesign

Revision ID: 239cea6d2452
Revises: 140c971ae0a3
Create Date: 2018-12-04 21:14:15.250247

"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from alembic import op
from sqlalchemy import String, Integer
from sqlalchemy.sql import table, column, select, text
from sqlalchemy.dialects.postgresql import UUID

# revision identifiers, used by Alembic.
revision = '239cea6d2452'
down_revision = '140c971ae0a3'
branch_labels = None
depends_on = None


def migrate_infer_calculation_entry_point(connection):
    """Set the process type for calculation nodes by inferring it from their type string."""
    from aiida.manage.database.integrity import write_database_integrity_violation
    from aiida.manage.database.integrity.plugins import infer_calculation_entry_point
    from aiida.plugins.entry_point import ENTRY_POINT_STRING_SEPARATOR

    DbNode = table('db_dbnode', column('id', Integer), column('uuid', UUID), column('type', String),
                   column('process_type', String))

    query_set = connection.execute(select([DbNode.c.type]).where(DbNode.c.type.like('calculation.%'))).fetchall()
    type_strings = set(entry[0] for entry in query_set)
    mapping_node_type_to_entry_point = infer_calculation_entry_point(type_strings=type_strings)

    fallback_cases = []

    for type_string, entry_point_string in mapping_node_type_to_entry_point.items():

        # If the entry point string does not contain the entry point string separator, the mapping function was not able
        # to map the type string onto a known entry point string. As a fallback it uses the modified type string itself.
        # All affected entries should be logged to file that the user can consult.
        if ENTRY_POINT_STRING_SEPARATOR not in entry_point_string:
            query_set = connection.execute(
                select([DbNode.c.uuid]).where(DbNode.c.type == op.inline_literal(type_string))).fetchall()

            uuids = [str(entry.uuid) for entry in query_set]
            for uuid in uuids:
                fallback_cases.append([uuid, type_string, entry_point_string])

        connection.execute(DbNode.update().where(DbNode.c.type == op.inline_literal(type_string)).values(
            process_type=op.inline_literal(entry_point_string)))

    if fallback_cases:
        headers = ['UUID', 'type (old)', 'process_type (fallback)']
        warning_message = 'found calculation nodes with a type string that could not be mapped onto a known entry point'
        action_message = 'inferred `process_type` for all calculation nodes, using fallback for unknown entry points'
        write_database_integrity_violation(fallback_cases, headers, warning_message, action_message)


def detect_unexpected_links(connection):
    """Scan the database for any links that are unexpected.

    The checks will verify that there are no outgoing `call` or `return` links from calculation nodes and that if a
    workflow node has a `create` link, it has at least an accompanying return link to the same data node, or it has a
    `call` link to a calculation node that takes the created data node as input.
    """
    from aiida.backends.general.migrations.provenance_redesign import INVALID_LINK_SELECT_STATEMENTS
    from aiida.manage.database.integrity import write_database_integrity_violation

    for sql, warning_message in INVALID_LINK_SELECT_STATEMENTS:
        results = list(connection.execute(text(sql)))
        if results:
            headers = ['UUID source', 'UUID target', 'link type', 'link label']
            write_database_integrity_violation(results, headers, warning_message)


def upgrade():
    """The upgrade migration actions."""
    connection = op.get_bind()

    # Migrate calculation nodes by inferring the process type from the type string
    migrate_infer_calculation_entry_point(connection)

    # Detect if the database contain any unexpected links
    detect_unexpected_links(connection)

    statement = text("""
        DELETE FROM db_dblink WHERE db_dblink.id IN (
            SELECT db_dblink.id FROM db_dblink
            INNER JOIN db_dbnode ON db_dblink.input_id = db_dbnode.id
            WHERE
                (db_dbnode.type LIKE 'calculation.job%' OR db_dbnode.type LIKE 'calculation.inline%')
                AND db_dblink.type = 'returnlink'
        ); -- Delete all outgoing RETURN links from JobCalculation and InlineCalculation nodes

        DELETE FROM db_dblink WHERE db_dblink.id IN (
            SELECT db_dblink.id FROM db_dblink
            INNER JOIN db_dbnode ON db_dblink.input_id = db_dbnode.id
            WHERE
                (db_dbnode.type LIKE 'calculation.job%' OR db_dbnode.type LIKE 'calculation.inline%')
                AND db_dblink.type = 'calllink'
        ); -- Delete all outgoing CALL links from JobCalculation and InlineCalculation nodes

        DELETE FROM db_dblink WHERE db_dblink.id IN (
            SELECT db_dblink.id FROM db_dblink
            INNER JOIN db_dbnode ON db_dblink.input_id = db_dbnode.id
            WHERE
                (db_dbnode.type LIKE 'calculation.function%' OR db_dbnode.type LIKE 'calculation.work%')
                AND db_dblink.type = 'createlink'
        ); -- Delete all outgoing CREATE links from WorkCalculation nodes

        UPDATE db_dbnode SET type = 'calculation.work.WorkCalculation.'
        WHERE type = 'calculation.process.ProcessCalculation.';
         -- First migrate very old `ProcessCalculation` to `WorkCalculation`

        UPDATE db_dbnode SET type = 'node.process.workflow.workfunction.WorkFunctionNode.'
        WHERE type = 'calculation.work.WorkCalculation.'
            AND attributes ? 'function_name';
         -- WorkCalculations that have a `function_name` attribute are `WorkFunctionNode`

        UPDATE db_dbnode SET type = 'node.process.workflow.workchain.WorkChainNode.'
        WHERE type = 'calculation.work.WorkCalculation.';
         -- Update type for `WorkCalculation` nodes - all what is left should be `WorkChainNodes`

        UPDATE db_dbnode SET type = 'node.process.calculation.calcjob.CalcJobNode.'
        WHERE type LIKE 'calculation.job.%'; -- Update type for JobCalculation nodes

        UPDATE db_dbnode SET type = 'node.process.calculation.calcfunction.CalcFunctionNode.'
        WHERE type = 'calculation.inline.InlineCalculation.'; -- Update type for InlineCalculation nodes

        UPDATE db_dbnode SET type = 'node.process.workflow.workfunction.WorkFunctionNode.'
        WHERE type = 'calculation.function.FunctionCalculation.'; -- Update type for FunctionCalculation nodes

        UPDATE db_dblink SET type = 'create' WHERE type = 'createlink'; -- Rename `createlink` to `create`
        UPDATE db_dblink SET type = 'return' WHERE type = 'returnlink'; -- Rename `returnlink` to `return`

        UPDATE db_dblink SET type = 'input_calc' FROM db_dbnode
        WHERE db_dblink.output_id = db_dbnode.id AND db_dbnode.type LIKE 'node.process.calculation%'
        AND db_dblink.type = 'inputlink';
         -- Rename `inputlink` to `input_calc` if the target node is a calculation type node

        UPDATE db_dblink SET type = 'input_work' FROM db_dbnode
        WHERE db_dblink.output_id = db_dbnode.id AND db_dbnode.type LIKE 'node.process.workflow%'
        AND db_dblink.type = 'inputlink';
         -- Rename `inputlink` to `input_work` if the target node is a workflow type node

        UPDATE db_dblink SET type = 'call_calc' FROM db_dbnode
        WHERE db_dblink.output_id = db_dbnode.id AND db_dbnode.type LIKE 'node.process.calculation%'
        AND db_dblink.type = 'calllink';
         -- Rename `calllink` to `call_calc` if the target node is a calculation type node

        UPDATE db_dblink SET type = 'call_work' FROM db_dbnode
        WHERE db_dblink.output_id = db_dbnode.id AND db_dbnode.type LIKE 'node.process.workflow%'
        AND db_dblink.type = 'calllink';
         -- Rename `calllink` to `call_work` if the target node is a workflow type node
        """)
    connection.execute(statement)


def downgrade():
    """The downgrade migration actions."""
    connection = op.get_bind()

    statement = text("""
        UPDATE db_dbnode SET type = 'calculation.job.JobCalculation.'
        WHERE type = 'node.process.calculation.calcjob.CalcJobNode.';

        UPDATE db_dbnode SET type = 'calculatison.inline.InlineCalculation.'
        WHERE type = 'node.process.calculation.calcfunction.CalcFunctionNode.';

        UPDATE db_dbnode SET type = 'calculation.function.FunctionCalculation.'
        WHERE type = 'node.process.workflow.workfunction.WorkFunctionNode.';

        UPDATE db_dbnode SET type = 'calculation.work.WorkCalculation.'
        WHERE type = 'node.process.workflow.workchain.WorkChainNode.';


        UPDATE db_dblink SET type = 'inputlink'
        WHERE type = 'input_call' OR type = 'input_work';

        UPDATE db_dblink SET type = 'calllink'
        WHERE type = 'call_call' OR type = 'call_work';

        UPDATE db_dblink SET type = 'createlink'
        WHERE type = 'create';

        UPDATE db_dblink SET type = 'returnlink'
        WHERE type = 'return';
        """)
    connection.execute(statement)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module provides import functionality to all data types
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import io

from aiida.cmdline.utils import echo


def data_import_xyz(filename, **kwargs):
    """
    Imports an XYZ-file.
    """
    from os.path import abspath
    from aiida.orm import StructureData

    vacuum_addition = kwargs.pop('vacuum_addition')
    vacuum_factor = kwargs.pop('vacuum_factor')
    pbc = [bool(i) for i in kwargs.pop('pbc')]
    store = kwargs.pop('store')
    view_in_ase = kwargs.pop('view')

    echo.echo('importing XYZ-structure from: \n  {}'.format(abspath(filename)))
    filepath = abspath(filename)
    with io.open(filepath, encoding='utf8') as fobj:
        xyz_txt = fobj.read()
    new_structure = StructureData()
    # pylint: disable=protected-access
    try:
        new_structure._parse_xyz(xyz_txt)
        new_structure._adjust_default_cell(vacuum_addition=vacuum_addition, vacuum_factor=vacuum_factor, pbc=pbc)

        if store:
            new_structure.store()
        if view_in_ase:
            from ase.visualize import view
            view(new_structure.get_ase())
        echo.echo('  Succesfully imported structure {}, '
                  '(PK = {})'.format(new_structure.get_formula(), new_structure.pk))

    except ValueError as err:
        echo.echo_critical(err)


def data_import_pwi(filename, **kwargs):
    """
    Imports a structure from a quantumespresso input file.
    """
    from os.path import abspath
    try:
        from qe_tools.parsers.pwinputparser import PwInputFile
    except ImportError:
        echo.echo_critical("You have not installed the package qe-tools. \n"
                           "You can install it with: pip install qe-tools")

    store = kwargs.pop('store')
    view_in_ase = kwargs.pop('view')

    echo.echo('importing structure from: \n  {}'.format(abspath(filename)))
    filepath = abspath(filename)

    try:
        inputparser = PwInputFile(filepath)
        new_structure = inputparser.get_structuredata()

        if store:
            new_structure.store()
        if view_in_ase:
            from ase.visualize import view
            view(new_structure.get_ase())
        echo.echo('  Succesfully imported structure {}, '
                  '(PK = {})'.format(new_structure.get_formula(), new_structure.pk))

    except ValueError as err:
        echo.echo_critical(err)


def data_import_ase(filename, **kwargs):
    """
    Imports a structure in a number of formats using the ASE routines.
    """
    from os.path import abspath
    from aiida.orm import StructureData

    try:
        import ase.io
    except ImportError:
        echo.echo_critical("You have not installed the package ase. \n" "You can install it with: pip install ase")

    store = kwargs.pop('store')
    view_in_ase = kwargs.pop('view')

    echo.echo('importing structure from: \n  {}'.format(abspath(filename)))
    filepath = abspath(filename)

    try:
        asecell = ase.io.read(filepath)
        new_structure = StructureData(ase=asecell)

        if store:
            new_structure.store()
        if view_in_ase:
            from ase.visualize import view
            view(new_structure.get_ase())
        echo.echo('  Succesfully imported structure {}, '
                  '(PK = {})'.format(new_structure.get_formula(), new_structure.pk))

    except ValueError as err:
        echo.echo_critical(err)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi import`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os

import unittest
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_import


def get_archive_file(archive):
    """
    Return the absolute path of the archive file used for testing purposes. The expected base path for these files:

        aiida.backends.tests.fixtures

    :param archive: the relative filename of the archive
    :returns: absolute filepath of the archive test file
    """
    dirpath_current = os.path.dirname(os.path.abspath(__file__))
    dirpath_archive = os.path.join(dirpath_current, os.pardir, os.pardir, 'fixtures')

    return os.path.join(dirpath_archive, archive)


class TestVerdiImport(AiidaTestCase):
    """Tests for `verdi import`."""

    def setUp(self):
        self.cli_runner = CliRunner()

    def test_import_no_archives(self):
        """Test that passing no valid archives will lead to command failure."""
        options = []
        result = self.cli_runner.invoke(cmd_import.cmd_import, options)

        self.assertIsNotNone(result.exception)
        self.assertIn('Critical', result.output)

    def test_import_non_existing_archives(self):
        """Test that passing a non-existing archive will lead to command failure."""
        options = ['non-existing-archive.aiida']
        result = self.cli_runner.invoke(cmd_import.cmd_import, options)

        self.assertIsNotNone(result.exception)

    @unittest.skip("Reenable when issue #2426 has been solved (migrate exported files from 0.3 to 0.4)")
    def test_import_archive(self):
        """
        Test import for archive files from disk

        NOTE: When the export format version is upped, the test export_v0.4.aiida archive will have to be
        replaced with the version of the new format
        """
        archives = [
            get_archive_file('calcjob/arithmetic.add.aiida'),
            get_archive_file('export/migrate/export_v0.4.aiida')
        ]

        options = [] + archives
        result = self.cli_runner.invoke(cmd_import.cmd_import, options)

        self.assertIsNone(result.exception, result.output)

    @unittest.skip("Reenable when issue #2426 has been solved (migrate exported files from 0.3 to 0.4)")
    def test_comment_mode(self):
        """
        Test comment mode flag works as intended
        """
        archives = [get_archive_file('export/migrate/export_v0.4.aiida')]

        options = ['--comment-mode', 'newest'] + archives
        result = self.cli_runner.invoke(cmd_import.cmd_import, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Comment mode: newest', result.output)

        options = ['--comment-mode', 'overwrite'] + archives
        result = self.cli_runner.invoke(cmd_import.cmd_import, options)
        self.assertIsNone(result.exception, result.output)
        self.assertIn('Comment mode: overwrite', result.output)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test utility to import fixtures, such as export archives."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os


def import_archive_fixture(filepath):
    """Import a test fixture that is an AiiDA export archive

    :param filepath: the relative path of the archive file within the fixture directory
    """
    from aiida.orm.importexport import import_data

    filepath_current = os.path.dirname(os.path.realpath(__file__))
    filepath_fixtures = os.path.join(filepath_current, os.pardir, 'fixtures')
    filepath_archive = os.path.join(filepath_fixtures, filepath)

    if not os.path.isfile(filepath_archive):
        raise ValueError('archive {} does not exist in the fixture directory {}'.format(filepath, filepath_fixtures))

    import_data(filepath_archive, silent=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Django implementations for the `AuthInfo` entity and collection."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.djsite.db.models import DbAuthInfo
from aiida.common import exceptions
from aiida.common import json
from aiida.common.lang import type_check

from ..authinfos import BackendAuthInfo, BackendAuthInfoCollection
from . import entities
from . import utils


class DjangoAuthInfo(entities.DjangoModelEntity[DbAuthInfo], BackendAuthInfo):
    """AuthInfo implementation for Django."""

    MODEL_CLASS = DbAuthInfo

    @classmethod
    def get_dbmodel_attribute_name(cls, attr_name):
        """Return the name of the auth info attribute as it is known to the database model.

        This is essentially a mapping because the `type_string` attribute is called `type` on the database model class.

        :return: name of the backend `attribute` as defined on the database model class
        """
        if attr_name == 'type_string':
            return 'type'

        return super(DjangoAuthInfo, cls).get_dbmodel_attribute_name(attr_name)

    def __init__(self, backend, computer, user):
        """
        Construct a DjangoAuthInfo.

        :param computer: a Computer instance
        :param user: a User instance
        :return: an AuthInfo object associated with the given computer and user
        """
        from . import computers
        from . import users
        super(DjangoAuthInfo, self).__init__(backend)
        type_check(user, users.DjangoUser)
        type_check(computer, computers.DjangoComputer)
        self._dbmodel = utils.ModelWrapper(DbAuthInfo(dbcomputer=computer.dbmodel, aiidauser=user.dbmodel))

    @property
    def is_stored(self):
        """Return whether the `AuthInfo` is stored

        :return: True if stored, False otherwise
        """
        return self._dbmodel.is_saved()

    @property
    def id(self):
        return self._dbmodel.id

    @property
    def enabled(self):
        return self._dbmodel.enabled

    @enabled.setter
    def enabled(self, enabled):
        self._dbmodel.enabled = enabled

    @property
    def computer(self):
        return self.backend.computers.from_dbmodel(self._dbmodel.dbcomputer)

    @property
    def user(self):
        return self._backend.users.from_dbmodel(self._dbmodel.aiidauser)

    def get_auth_params(self):
        """
        Get the auth_params dictionary from the DB

        :return: a dictionary
        """
        try:
            return json.loads(self._dbmodel.auth_params)
        except ValueError:
            email = self._dbmodel.aiidauser.email
            hostname = self._dbmodel.dbcomputer.hostname
            raise exceptions.DbContentError(
                "Error while reading auth_params for dbauthinfo, aiidauser={}, computer={}".format(email, hostname))

    def set_auth_params(self, auth_params):
        """
        Replace the auth_params dictionary in the DB with the provided dictionary
        """
        # Raises ValueError if data is not JSON-serializable
        self._dbmodel.auth_params = json.dumps(auth_params)

    def get_metadata(self):
        """
        Get the metadata dictionary from the DB

        :return: a dictionary
        """
        try:
            return json.loads(self._dbmodel.metadata)
        except ValueError:
            raise exceptions.DbContentError(
                "Error while reading metadata for dbauthinfo, aiidauser={}, computer={}".format(
                    self.aiidauser.email, self.dbcomputer.hostname))

    def set_metadata(self, metadata):
        """
        Replace the metadata dictionary in the DB with the provided dictionary
        """
        # Raises ValueError if data is not JSON-serializable
        self._dbmodel.metadata = json.dumps(metadata)

    def store(self):
        """
        Store the AuthInfo (possibly updating values if changed)

        :return: the AuthInfo instance
        """
        self._dbmodel.save()
        return self


class DjangoAuthInfoCollection(BackendAuthInfoCollection):
    """Collection of AuthInfo instances."""

    ENTITY_CLASS = DjangoAuthInfo

    def get(self, computer, user):
        """
        Return a AuthInfo given a computer and a user

        :param computer: a Computer instance
        :param user: a User instance
        :return: an AuthInfo object associated with the given computer and user
        :raise aiida.common.NotExistent: if the user is not configured to use computer
        :raise sqlalchemy.orm.exc.MultipleResultsFound: if the user is configured
            more than once to use the computer! Should never happen
        """
        # pylint: disable=import-error,no-name-in-module
        from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

        try:
            authinfo = DbAuthInfo.objects.get(dbcomputer=computer.id, aiidauser=user.id)
            return self.from_dbmodel(authinfo)
        except ObjectDoesNotExist:
            raise exceptions.NotExistent("The aiida user {} is not configured to use computer {}".format(
                user.email, computer.name))
        except MultipleObjectsReturned:
            raise exceptions.ConfigurationError("The aiida user {} is configured more than once to use "
                                                "computer {}! Only one configuration is allowed".format(
                                                    user.email, computer.name))

    def delete(self, authinfo_id):
        """Delete an `AuthInfo` entry from the collection and database.

        :param authinfo_id: the id of the entity
        """
        # pylint: disable=import-error,no-name-in-module
        from django.core.exceptions import ObjectDoesNotExist
        try:
            DbAuthInfo.objects.get(pk=authinfo_id).delete()
        except ObjectDoesNotExist:
            raise exceptions.NotExistent("AuthInfo with id '{}' not found".format(authinfo_id))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Comment objects and functions"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.manage.manager import get_manager
from . import entities
from . import users

__all__ = ('Comment',)


class Comment(entities.Entity):
    """Base class to map a DbComment that represents a comment attached to a certain Node."""

    class Collection(entities.Collection):
        """The collection of Comment entries."""

        def delete(self, comment_id):
            """
            Remove a Comment from the collection with the given id

            :param comment_id: the id of the comment to delete
            """
            self._backend.comments.delete(comment_id)

    def __init__(self, node, user, content=None, backend=None):
        """
        Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :return: a Comment object associated to the given node and user
        """
        backend = backend or get_manager().get_backend()
        model = backend.comments.create(node=node.backend_entity, user=user.backend_entity, content=content)
        super(Comment, self).__init__(model)

    def __str__(self):
        arguments = [self.uuid, self.node.pk, self.user.email, self.content]
        return 'Comment<{}> for node<{}> and user<{}>: {}'.format(*arguments)

    @property
    def ctime(self):
        return self._backend_entity.ctime

    @property
    def mtime(self):
        return self._backend_entity.mtime

    def set_mtime(self, value):
        return self._backend_entity.set_mtime(value)

    @property
    def node(self):
        return self._backend_entity.node

    @property
    def user(self):
        return users.User.from_backend_entity(self._backend_entity.user)

    def set_user(self, value):
        self._backend_entity.user = value.backend_entity

    @property
    def content(self):
        return self._backend_entity.content

    def set_content(self, value):
        return self._backend_entity.set_content(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Generic backend related objects"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import typing
import six

__all__ = ('Backend', 'BackendEntity', 'BackendCollection', 'EntityType')

EntityType = typing.TypeVar('EntityType')  # pylint: disable=invalid-name


@six.add_metaclass(abc.ABCMeta)
class Backend(object):
    """The public interface that defines a backend factory that creates backend specific concrete objects."""

    # pylint: disable=useless-object-inheritance

    @abc.abstractmethod
    def migrate(self):
        """Migrate the database to the latest schema version."""

    @abc.abstractproperty
    def authinfos(self):
        """
        Return the collection of authorisation information objects

        :return: the authinfo collection
        :rtype: :class:`aiida.orm.implementation.BackendAuthInfoCollection`
        """

    @abc.abstractproperty
    def comments(self):
        """
        Return the collection of comments

        :return: the comment collection
        :rtype: :class:`aiida.orm.implementation.BackendCommentCollection`
        """

    @abc.abstractproperty
    def computers(self):
        """
        Return the collection of computers

        :return: the computers collection
        :rtype: :class:`aiida.orm.implementation.BackendComputerCollection`
        """

    @abc.abstractproperty
    def groups(self):
        """
        Return the collection of groups

        :return: the groups collection
        :rtype: :class:`aiida.orm.implementation.BackendGroupCollection`
        """

    @abc.abstractproperty
    def logs(self):
        """
        Return the collection of logs

        :return: the log collection
        :rtype: :class:`aiida.orm.implementation.BackendLogCollection`
        """

    @abc.abstractproperty
    def nodes(self):
        """
        Return the collection of nodes

        :return: the nodes collection
        :rtype: :class:`aiida.orm.implementation.BackendNodeCollection`
        """

    @abc.abstractproperty
    def query_manager(self):
        """
        Return the query manager for the objects stored in the backend

        :return: The query manger
        :rtype: :class:`aiida.backends.general.abstractqueries.AbstractQueryManager`
        """

    @abc.abstractmethod
    def query(self):
        """
        Return an instance of a query builder implementation for this backend

        :return: a new query builder instance
        :rtype: :class:`aiida.orm.implementation.BackendQueryBuilder`
        """

    @abc.abstractproperty
    def users(self):
        """
        Return the collection of users

        :return: the users collection
        :rtype: :class:`aiida.orm.implementation.BackendUserCollection`
        """

    @abc.abstractmethod
    def transaction(self):
        """
        Get a context manager that can be used as a transaction context for a series of backend operations.
        If there is an exception within the context then the changes will be rolled back and the state will
        be as before entering.  Transactions can be nested.

        :return: a context manager to group database operations
        """


@six.add_metaclass(abc.ABCMeta)
class BackendEntity(object):
    """An first-class entity in the backend"""

    # pylint: disable=useless-object-inheritance

    def __init__(self, backend):
        self._backend = backend
        self._dbmodel = None

    @property
    def backend(self):
        """Return the backend this entity belongs to

        :return: the backend instance
        """
        return self._backend

    @property
    def dbmodel(self):
        return self._dbmodel

    @abc.abstractproperty
    def id(self):  # pylint: disable=invalid-name
        """Return the id for this entity.

        This is unique only amongst entities of this type for a particular backend.

        :return: the entity id
        """

    @property
    def pk(self):
        """Return the id for this entity.

        This is unique only amongst entities of this type for a particular backend.

        :return: the entity id
        """
        return self.id

    @abc.abstractmethod
    def store(self):
        """Store this entity in the backend.

        Whether it is possible to call store more than once is delegated to the object itself
        """

    @abc.abstractproperty
    def is_stored(self):
        """Return whether the entity is stored.

        :return: True if stored, False otherwise
        :rtype: bool
        """


class BackendCollection(typing.Generic[EntityType]):
    """Container class that represents a collection of entries of a particular backend entity."""

    ENTITY_CLASS = None  # type: EntityType

    def __init__(self, backend):
        """
        :param backend: the backend this collection belongs to
        :type backend: :class:`aiida.orm.implementation.Backend`
        """
        assert issubclass(self.ENTITY_CLASS, BackendEntity), 'Must set the ENTRY_CLASS class variable to an entity type'
        self._backend = backend

    def from_dbmodel(self, dbmodel):
        """
        Create an entity from the backend dbmodel

        :param dbmodel: the dbmodel to create the entity from
        :return: the entity instance
        """
        return self.ENTITY_CLASS.from_dbmodel(dbmodel, self.backend)

    @property
    def backend(self):
        """
        Return the backend.

        :rtype: :class:`aiida.orm.implementation.Backend`
        """
        return self._backend

    def create(self, **kwargs):
        """
        Create new a entry and set the attributes to those specified in the keyword arguments

        :return: the newly created entry of type ENTITY_CLASS
        """
        return self.ENTITY_CLASS(backend=self._backend, **kwargs)  # pylint: disable=not-callable
